/*********************************************************************NVMH1****
File:
dds.c

Copyright (C) 1999, 2000 NVIDIA Corporation
This file is provided without support, instruction, or implied warranty of any
kind.  NVIDIA makes no guarantee of its fitness for a particular purpose and is
not liable under any circumstances for any damages or loss whatsoever arising
from the use or inability to use this file or items derived from it.

Comments: Shows how to read and download compressed data images to GL

  Comments/Ideas can be sent to sebastien.domine@nvidia.com

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/glut.h>
#include <GL/glext.h>

#include "tga.h"

#include "dds.h"

/* needed for DirectX's DDSURFACEDESC2 structure definition */
#include <ddraw.h>

gliGenericImage *
ReadDDSFile(const char *filename, int * bufsize, int * numMipmaps) 
{
  gliGenericImage *genericImage;
  DDSURFACEDESC2 ddsd;
  char filecode[4];
  FILE *fp;
  int factor;

  /* try to open the file */
  fp = fopen(filename, "rb");
  if (fp == NULL)
    return NULL;
  
  /* verify the type of file */
  fread(filecode, 1, 4, fp);
  if (strncmp(filecode, "DDS ", 4) != 0) {
    fclose(fp);
    return NULL;
  }
   
  /* get the surface desc */
  fread(&ddsd, sizeof(ddsd), 1, fp);

  genericImage = (gliGenericImage*) malloc(sizeof(gliGenericImage));
  memset(genericImage,0,sizeof(gliGenericImage));

  switch(ddsd.ddpfPixelFormat.dwFourCC)
  {
    case FOURCC_DXT1:
      genericImage->format = GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
      factor = 2;
      break;
    case FOURCC_DXT3:
      genericImage->format = GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
      factor = 4;
      break;
    case FOURCC_DXT5:
      genericImage->format = GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
      factor = 4;
      break;
    default:
      return NULL;
  }

  /* how big is it going to be including all mipmaps? */
  *bufsize = ddsd.dwMipMapCount > 1 ? ddsd.dwLinearSize * factor : ddsd.dwLinearSize;
  genericImage->pixels = (unsigned char*)malloc(*bufsize * sizeof(unsigned char));
  fread(genericImage->pixels, 1, *bufsize, fp);
  /* close the file pointer */
  fclose(fp);
  
  genericImage->width       = ddsd.dwWidth;
  genericImage->height      = ddsd.dwHeight;
  genericImage->components  = (ddsd.ddpfPixelFormat.dwFourCC == FOURCC_DXT1) ? 3 : 4;
  *numMipmaps = ddsd.dwMipMapCount;

  /* return data */
  return genericImage;
}
